/*
 * StrawLib v0.2
 *
 * Copyright (c) 2005 Pr Spjuth
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, publish, distribute, 
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or 
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef STRAW_H_INCLUDED
#define STRAW_H_INCLUDED


/**************************************************************************************************
 * Configuration
 **************************************************************************************************/

/**
 * 
 */
#define STRAW_CFG_DEBUG


/**************************************************************************************************
 * StrawLib API Macros
 **************************************************************************************************/
/* NOTE: This is the only public part of the StrawLib, all other types and functions are to be
 *       considered private and may change in future versions. Always use the macros!
 */

/**
 * 
 */
#define STRAW_SESSION  StrawSession_t


/**
 * 
 */
/* NOTE: This operation must be split up like this because the Straw class concept depends
 *       solely on how the preprocessor have created functions to start child-straws.
 */
#define STRAW_CREATESESSION(Session_p, AllocMem, FreeMem, ClassName)                               \
  {                                                                                                \
    StrawSession_t *CurrentSession_p = (Session_p);                                                \
    StrawCreateSession((Session_p), AllocMem, FreeMem);                                            \
    (Session_p)->StrawList_p = CHUNK_STARTCHILDSTRAW(ClassName, NULL);                             \
  }


/**
 * 
 */
#define STRAW_SESSIONPROCESS(Session_p, InternalSignal, ExternalSignal, SignalData_p,              \
                             SignalConsumed_p)                                                     \
  StrawSessionProcess(Session_p, InternalSignal, ExternalSignal, SignalData_p, SignalConsumed_p)


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  #define STRAW_DECLARECLASS(ClassName)                                                            \
    Straw_t*                                                                                       \
    StrawClass##ClassName(StrawSession_t*              CurrentSession_p,                           \
                          const PrivateDataType* const PrivateData_p,                              \
                          StrawChunk_t                 ReturnChunk,                                \
                          const char*                  ReturnLabel_p)
#else
  #define STRAW_DECLARECLASS(ClassName)                                                            \
    Straw_t*                                                                                       \
    StrawClass##ClassName(StrawSession_t*              CurrentSession_p,                           \
                          const PrivateDataType* const PrivateData_p,                              \
                          StrawChunk_t                 ReturnChunk)
#endif


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  #define STRAW_DECLARECLASSNODATA(ClassName)                                                      \
    Straw_t*                                                                                       \
    StrawClass##ClassName(StrawSession_t*   CurrentSession_p,                                      \
                          const void* const PrivateData_p,                                         \
                          StrawChunk_t      ReturnChunk,                                           \
                          const char*       ReturnLabel_p)
#else
  #define STRAW_DECLARECLASSNODATA(ClassName)                                                      \
    Straw_t*                                                                                       \
    StrawClass##ClassName(StrawSession_t*   CurrentSession_p,                                      \
                          const void* const PrivateData_p,                                         \
                          StrawChunk_t      ReturnChunk)
#endif


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  #define STRAW_DEFINECLASS(ClassName, EntryChunk, PrivateDataType)                                \
    Straw_t*                                                                                       \
    StrawClass##ClassName(StrawSession_t*              CurrentSession_p,                           \
                          const PrivateDataType* const PrivateData_p,                              \
                          StrawChunk_t                 ReturnChunk,                                \
                          const char*                  ReturnLabel_p)                              \
    {                                                                                              \
      return StrawStartChild(CurrentSession_p, EntryChunk, ReturnChunk, ReturnLabel_p,             \
                             PrivateData_p, sizeof(PrivateDataType));                              \
    }                                                                                              \
    typedef int StrawDummyType##StrawClass
#else
  #define STRAW_DEFINECLASS(ClassName, EntryChunk, PrivateDataType)                                \
    Straw_t*                                                                                       \
    StrawClass##ClassName(StrawSession_t*              CurrentSession_p,                           \
                          const PrivateDataType* const PrivateData_p,                              \
                          StrawChunk_t                 ReturnChunk)                                \
    {                                                                                              \
      return StrawStartChild(CurrentSession_p, EntryChunk, ReturnChunk, PrivateData_p,             \
                             sizeof(PrivateDataType));                                             \
    }                                                                                              \
    typedef int StrawDummyType##StrawClass
#endif


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  #define STRAW_DEFINECLASSNODATA(ClassName, EntryChunk)                                           \
    Straw_t*                                                                                       \
    StrawClass##ClassName(StrawSession_t*   CurrentSession_p,                                      \
                          const void* const PrivateData_p,                                         \
                          StrawChunk_t      ReturnChunk,                                           \
                          const char*       ReturnLabel_p)                                         \
    {                                                                                              \
      (void)PrivateData_p;                                                                         \
      return StrawStartChild(CurrentSession_p, EntryChunk, ReturnChunk, ReturnLabel_p, NULL, 0);   \
    }                                                                                              \
    typedef int StrawDummyType##StrawClass
#else
  #define STRAW_DEFINECLASSNODATA(ClassName, EntryChunk)                                           \
    Straw_t*                                                                                       \
    StrawClass##ClassName(StrawSession_t*   CurrentSession_p,                                      \
                          const void* const PrivateData_p,                                         \
                          StrawChunk_t      ReturnChunk)                                           \
    {                                                                                              \
      (void)PrivateData_p;                                                                         \
      return StrawStartChild(CurrentSession_p, EntryChunk, ReturnChunk, NULL, 0);                  \
    }                                                                                              \
    typedef int StrawDummyType##StrawClass
#endif


/**
 * 
 */
// TODO: Make this variable name something more unique to avoid conflicts! (Prepend "Straw")
#define CHUNK_PARAMS  StrawSession_t* CurrentSession_p


/**
 * 
 */
#define CHUNK_MATCHALL  StrawMatchAll


/**
 * 
 */
#define CHUNK_MATCHDROPPED  StrawMatchDropped


/**
 * 
 */
#define CHUNK_GETSIGNALTYPE()                                                                      \
  (CurrentSession_p->SignalType)


/**
 * 
 */
#define CHUNK_GETSIGNALDATA(DataType)                                                              \
  ((*DataType)CurrentSession_p->SignalData_p)


/**
 * 
 */
#define CHUNK_GETSTRAWID()                                                                         \
  ((unsigned int)CurrentSession_p->CurrentStraw_p)


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  #define CHUNK_SETCLEANUP(ChunkName)                                                              \
    StrawLogSetCleanup(CurrentSession_p->CurrentStraw_p, ChunkName == NULL, #ChunkName);           \
    if ((CurrentSession_p->CurrentStraw_p->Flags & STRAW_TERMINATING) > 0) {                       \
      CurrentSession_p->CurrentStraw_p->SlotA.CleanupChunk = NULL;                                 \
      CurrentSession_p->CurrentStraw_p->CleanupLabel_p = NULL;                                     \
    }                                                                                              \
    else {                                                                                         \
      CurrentSession_p->CurrentStraw_p->SlotA.CleanupChunk = ChunkName;                            \
      CurrentSession_p->CurrentStraw_p->CleanupLabel_p = #ChunkName;                               \
    }  
#else
  #define CHUNK_SETCLEANUP(ChunkName)                                                              \
    CurrentSession_p->CurrentStraw_p->SlotA.CleanupChunk =                                         \
      ((CurrentSession_p->CurrentStraw_p->Flags & STRAW_TERMINATING) > 0)?NULL:ChunkName
#endif


/**
 * 
 */
#define CHUNK_JUMP(ChunkName)                                                                      \
  StrawLogJump(CurrentSession_p->CurrentStraw_p, #ChunkName);                                      \
  CurrentSession_p->CurrentStraw_p->SlotB.CurrentChunk = ChunkName;                                \
  return CHUNK_NEXTCHUNK


/**
 * 
 */
#define CHUNK_YIELD(ChunkName)                                                                     \
  StrawLogYield(CurrentSession_p->CurrentStraw_p, #ChunkName);                                     \
  CurrentSession_p->CurrentStraw_p->SlotB.CurrentChunk = ChunkName;                                \
  return CHUNK_YIELD


/**
 * 
 */
#define CHUNK_EXITSTRAW()                                                                          \
  return CHUNK_TERMINATE


/**
 * 
 */
#define CHUNK_TERMINATESTRAW(Straw, NotifyChunk)                                                   \
  StrawLogTerminateStraw(CurrentSession_p->CurrentStraw_p, Straw, #NotifyChunk);                   \
  if (StrawTerminate(CurrentSession_p, Straw, NotifyChunk)) {                                      \
    return CHUNK_WAITTERMINATION;                                                                  \
  }                                                                                                \
  else {                                                                                           \
    return CHUNK_NEXTCHUNK;                                                                        \
  }


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  #define CHUNK_INITNODATA(ChunkName)                                                              \
    StrawLogInit(CurrentSession_p->CurrentStraw_p, #ChunkName);                                    \
    CurrentSession_p->CurrentStraw_p->Label_p = #ChunkName
#else
  #define CHUNK_INITNODATA(ChunkName)
#endif


/**
 * 
 */
#define CHUNK_INIT(ChunkName, VariableName, DataType)                                              \
  DataType* VariableName = (DataType*)((unsigned char*)CurrentSession_p->CurrentStraw_p +          \
                           sizeof(Straw_t));                                                       \
  (void)VariableName;                                                                              \
  CHUNK_INITNODATA(ChunkName)


/**
 * 
 */
#define CHUNK_BEGINMAP(Value)                                                                      \
  {                                                                                                \
    static const StrawSignalMapping_t SignalMap[] = {


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  #define CHUNK_MAPSIGNAL(InternalSignal, ExternalSignal, Matcher, Chunk)                          \
        {InternalSignal, ExternalSignal, Chunk, Matcher, #InternalSignal, #ExternalSignal, #Chunk},
#else
  #define CHUNK_MAPSIGNAL(SignalType, Matcher, Chunk)                                              \
        {InternalSignal, ExternalSignal, Matcher, Chunk},
#endif


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  #define CHUNK_ENDMAPANDWAIT                                                                      \
        {0, 0, NULL, NULL, NULL, NULL, NULL}                                                       \
      };                                                                                           \
      StrawAddMapping(CurrentSession_p, (const StrawSignalMapping_t*)&SignalMap);                  \
      StrawLogWaitSignal(CurrentSession_p->CurrentStraw_p,                                         \
                         (const StrawSignalMapping_t*)&SignalMap);                                 \
    }                                                                                              \
    return CHUNK_WAITSIGNAL
#else
  #define CHUNK_ENDMAPANDWAIT                                                                      \
        {0, 0, 0, NULL, NULL}                                                                      \
      };                                                                                           \
      StrawAddMapping(CurrentSession_p, (const StrawSignalMapping_t*)&SignalMap);                  \
    }                                                                                              \
    return CHUNK_WAITSIGNAL
#endif


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  /* NOTE: The seemingly non-reuse way to write the CHUNK_MAPANDEXECUTE macro is done to preserve 
   *       signal definitions. Otherwise the preprocessor will expand their defines into integer
   *       constants before we can store the name for debugging purposes.
   */
  #define CHUNK_MAPANDWAIT(InternalSignal, ExternalSignal, Matcher, Chunk)                         \
    CHUNK_BEGINMAP(Value)                                                                          \
      {InternalSignal, ExternalSignal, Chunk, Matcher, #InternalSignal, #ExternalSignal, #Chunk},  \
    CHUNK_ENDMAPANDWAIT
#else
  #define CHUNK_MAPANDWAIT(InternalSignal, ExternalSignal, Matcher, Chunk)                         \
    CHUNK_BEGINMAP(Value)                                                                          \
      CHUNK_MAPSIGNAL(InternalSignal, ExternalSignal, Matcher, Chunk)                              \
    CHUNK_ENDMAPANDWAIT
#endif


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  #define CHUNK_RUNSUBSTRAW(ClassName, PrivateData_p, ReturnChunk)                                 \
    StrawClass##ClassName(CurrentSession_p, PrivateData_p, ReturnChunk, #ReturnChunk);             \
    return CHUNK_WAITSUBSTRAW
#else
  #define CHUNK_RUNSUBSTRAW(ClassName, PrivateData_p, ReturnChunk)                                 \
    StrawClass##ClassName(CurrentSession_p, PrivateData_p, ReturnChunk);                           \
    return CHUNK_WAITSUBSTRAW
#endif


/**
 * 
 */
#ifdef STRAW_CFG_DEBUG
  #define CHUNK_STARTCHILDSTRAW(ClassName, PrivateData_p)                                          \
    StrawClass##ClassName(CurrentSession_p, PrivateData_p, NULL, NULL);
#else
  #define CHUNK_STARTCHILDSTRAW(ClassName, PrivateData_p)                                          \
    StrawClass##ClassName(CurrentSession_p, PrivateData_p, NULL);
#endif


/**************************************************************************************************
 * Constants
 **************************************************************************************************/

/**
 * These are flags that indicates the current status of a straw. Some of the flags can be present
 * at the same time.
 *
 * @param  STRAW_TERMINATING         Straw is currently terminating, i.e. no more cleanup states
 *                                   should be set.
 * @param  STRAW_SUBSTRAW            Is a sub-straw that needs to notify its parent when exited.
 * @param  STRAW_PARENT              Has children which must be terminated before this parent is 
 *                                   terminated itself.
 * @param  STRAW_RUNNINGSUB          Currently waiting for a sub-straw to finish before continuing
 *                                   with the execution.
 * @param  STRAW_WAITFORSIGNAL       Waiting for a signal to arrive. This also means that there is
 *                                   a hash entry for this straw.
 * @param  STRAW_HASTERMINATOR       This straw has more straws that the parent that is interested
 *                                   in when it has terminated.
 * @param  STRAW_WAITFORTERMINATION  Currently waiting for another straw to terminate.
 * @param  STRAW_LASTTERMINATOR      Is the last terminator in a terminator list, which means that
 *                                   the TerminatedStraw field can be used.
 */
// TODO: Some of these flags cannot be present at the same time and should therefore share bits!
#define STRAW_TERMINATING        0x01
#define STRAW_SUBSTRAW           0x02
#define STRAW_PARENT             0x04
#define STRAW_RUNNINGSUB         0x08
#define STRAW_WAITFORSIGNAL      0x10
#define STRAW_HASTERMINATOR      0x20
#define STRAW_WAITFORTERMINATION 0x40
#define STRAW_LASTTERMINATOR     0x80


/**************************************************************************************************
 * Predefined Matching Functions
 **************************************************************************************************/

/**
 * Makes the signal mapping map to all signals of the given type. No specific matching operation
 * except for the signal type will be made.
 */
int
StrawMatchAll(void* StrawData_p, 
              void* SignalData_p);


/**
 * Makes the signal mapping map to all signals of the given type, if it was not matched by any other
 * straw. The only difference from StrawMatchAll() is that it lets all other straws have a peek
 * at the signal first.
 */
int
StrawMatchDropped(void* StrawData_p, 
                  void* SignalData_p);


/**************************************************************************************************
 * Type Definitions
 **************************************************************************************************/

/**
 * 
 */
typedef enum {
  STRAW_SESSIONACTIVE,
  STRAW_SESSIONWAITING,
  STRAW_SESSIONFINISHED
} StrawSessionStatus_t;


/**
 * 
 */
typedef enum {
  STRAW_TASKRUNSTRAW,
  STRAW_TASKTERMINATESTRAW,
  STRAW_TASKMAPSIGNAL
} StrawTaskAction_t;


/**
 * 
 */
typedef enum {
  CHUNK_WAITSIGNAL,
  CHUNK_WAITSUBSTRAW,
  CHUNK_WAITTERMINATION,
  CHUNK_NEXTCHUNK,
  CHUNK_YIELD,
  CHUNK_TERMINATE
} StrawChunkAction_t;


/**
 * 
 */
typedef void*
(*StrawAlloc_t)(size_t Size);


/**
 * 
 */
typedef void 
(*StrawFree_t)(void *Memory_p);


/**
 * 
 */
typedef int 
(*StrawSignalMatch_t)(void* StrawData_p, 
                      void* SignalData_p);


/**
 * 
 */
typedef struct StrawTask_st {
  StrawTaskAction_t    Action;
  void*                Data_p;
  struct StrawTask_st* NextTask_p;
  struct StrawTask_st* PrevTask_p; // TODO: If TERMINATION tasks are not prioritized this does not
                                   //       have to be present since we scan the list anyway on
                                   //       removal.
} StrawTask_t;


/**
 * 
 */
typedef struct StrawSession_st {
  StrawAlloc_t     AllocMem;
  StrawFree_t      FreeMem;
  StrawTask_t*     TaskQueueHead_p;
  StrawTask_t*     TaskQueueTail_p;
  struct Straw_st* StrawList_p;
  struct Straw_st* CurrentStraw_p;
  unsigned int     SignalType;
  void*            SignalData_p;
  struct Straw_st* MappingsHash_p[16];  
#ifdef STRAW_CFG_DEBUG
  unsigned long    Overhead;
  unsigned long    MaxOverhead;
#endif
} StrawSession_t;


/**
 * 
 */
typedef StrawChunkAction_t 
(*StrawChunk_t)(StrawSession_t *CurrentSession_p);


/**
 *
 * NOTE: This structure is always static why it does not, under normal circumstances, have to be
 *       optimized for memory usage.
 */
typedef struct {
  unsigned int       InternalSignal;
  unsigned int       ExternalSignal;
  StrawChunk_t       Chunk;
  StrawSignalMatch_t Matcher;
#ifdef STRAW_CFG_DEBUG
  char*              InternalName_p;
  char*              ExternalName_p;
  char*              ChunkName_p;
#endif
} StrawSignalMapping_t;


/**
 * 
 *
 * @param  PrivateData_p  
 * @param  ParentStraw_p  
 * @param  NextStraw_p    
 * @param  CleanupChunk      The chunk that should be executed when the straw is terminated, or NULL
 *                           if there is no need to run any chunk in that case.
 * @param  Terminator_p      Pointer to first straw that waits on this straw to terminate. This
 *                           field will be NULL if the straw was terminated as a consequence of the
 *                           parent terminating.
 * @param  CurrentChunk   
 * @param  NextHashEntry_p  
 * @param  NextTerminator_p
 * @param  SubStraw_p
 * @param  ChildCount
 * @param  SignalMapping_p
 * @param  Flags
 * @param  Label_p
 * @param  CleanupLabel_p
 */
/* NOTE: There are three slots (A, B and C) in the straw data that are used for data that do not
 *       need to be present at the same time. This is solely a RAM optimization and great care has
 *       to be taken when using and modifying the values in these slots!
 */
typedef struct Straw_st {
  struct Straw_st*              ParentStraw_p;
  struct Straw_st*              NextStraw_p;
  union {
    StrawChunk_t                CleanupChunk;
    struct Straw_st*            Terminator_p;
  } SlotA;
  union {
    StrawChunk_t                CurrentChunk;
    struct Straw_st*            NextHashEntry_p;
  } SlotB;
  union {
    struct Straw_st*            SubStraw_p;
    unsigned long               ChildCount;
    const StrawSignalMapping_t* SignalMapping_p;
    struct Straw_st*            NextTerminator_p;
    struct Straw_st*            TerminatedStraw_p;
  } SlotC;
  unsigned char                 Flags;
#ifdef STRAW_CFG_DEBUG
  char*                         Label_p;
  char*                         CleanupLabel_p;
#endif
} Straw_t;


/**************************************************************************************************
 * Framework Functions
 **************************************************************************************************/

/**
 * Initializes a straw framework session and creates a mother straw, starting at the
 * given entry point.
 */
void
StrawCreateSession(StrawSession_t* Session_p, 
                   StrawAlloc_t    AllocMem,
                   StrawFree_t     FreeMem);


/**
 * 
 */
StrawSessionStatus_t
StrawSessionProcess(StrawSession_t* Session_p,
                    unsigned int    InternalSignal,
                    unsigned int    ExternalSignal,
                    void*           SignalData_p,
                    int*            SignalConsumed_p);


/**
 * Starts a child- or sub-straw by creating all necessary data structures and scheduling the first
 * chunk in the straw class.
 *
 * @param  Session_p        
 * @param  EntryPoint_p     
 * @param  ReturnChunk_p    
 * @param  ReturnLabel_p    
 * @param  PrivateData_p    
 * @param  PrivateDataSize  
 *
 * @returns  A pointer to the new straw.
 */
Straw_t*
StrawStartChild(StrawSession_t* Session_p,
                StrawChunk_t    EntryPoint, 
                StrawChunk_t    ReturnChunk,
#ifdef STRAW_CFG_DEBUG
                const char*     ReturnLabel_p,
#endif
                const void*     PrivateData_p,
                unsigned int    PrivateDataSize);


/**
 * 
 * @returns  TRUE if the straw did exist and will now be terminated, FALSE otherwise.
 */
int
StrawTerminate(StrawSession_t* Session_p, 
               Straw_t*        Straw_p, 
               StrawChunk_t    ReturnChunk_p);


/**
 * 
 */
void
StrawAddMapping(StrawSession_t*             Session_p,
                const StrawSignalMapping_t* Mapping_p);


/************************************************************************************************
 * Logging Functions
 ************************************************************************************************/

#ifdef STRAW_CFG_DEBUG

/**
 * 
 */
void
StrawLogSetCleanup(Straw_t*    Straw_p,
                   int         Reset,
                   const char* CleanupChunk_p);


/**
 * 
 */
void
StrawLogJump(Straw_t*    Straw_p, 
             const char* NextChunk_p);


/**
 * 
 */
void
StrawLogYield(Straw_t*    Straw_p, 
              const char* NextChunk_p);


/**
 * 
 */
void
StrawLogTerminationAttempt(Straw_t* Straw_p);


/**
 * 
 */
void
StrawLogTerminateStraw(Straw_t*    Straw_p, 
                       Straw_t*    TerminatedStraw_p, 
                       const char* NotifyChunk_p);

/**
 * 
 */
void
StrawLogInit(Straw_t*    Straw_p,
             const char* ChunkName_p);


/**
 * 
 */
void
StrawLogWaitSignal(Straw_t*                    Straw_p,
                   const StrawSignalMapping_t* SignalMap_p);


/**
 * 
 */
void
StrawLogTerminateChildStraw(Straw_t* Straw_p,
                            Straw_t* ChildStraw_p,
                            int      SubStraw);


/**
 * 
 */
void
StrawLogExit(Straw_t* Straw_p);


/**
 * 
 */
void
StrawLogDeleted(Straw_t* Straw_p);


/**
 * 
 */
void
StrawLogRunCleanup(Straw_t*    Straw_p,
                   const char* CleanupChunk_p);


/**
 * 
 */
void
StrawLogRunChild(Straw_t* Straw_p,
                 Straw_t* NewStraw_p);


/**
 * 
 */
void
StrawLogRunSub(Straw_t*    Straw_p,
               Straw_t*    NewStraw_p,
               const char* ReturnChunk_p);


#endif /* STRAW_CFG_DEBUG */

#endif /* STRAW_H_INCLUDED */
