/*
 * StrawLib v0.2
 *
 * Copyright (c) 2005 Pr Spjuth
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, publish, distribute, 
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or 
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* StrawLib Headers */
#include "straws.h"

/* ANSI Library */
#include <stdio.h>


/************************************************************************************************
 * Global Functions
 ************************************************************************************************/

#ifdef STRAW_CFG_DEBUG

void
StrawLogSetCleanup(Straw_t*    Straw_p,
                   int         Reset,
                   const char* CleanupChunk_p)
{
  if (Reset) {
    printf("[%08X] \tCleanup chunk reset\n", Straw_p);
  }
  else if ((Straw_p->Flags & STRAW_TERMINATING) > 0) {
    printf("[%08X] \tNew cleanup chunk ignored - termination in progress\n", Straw_p);
  }
  else {
    printf("[%08X] \tCleanup chunk set to %s\n", Straw_p, CleanupChunk_p);
  }
}


void
StrawLogJump(Straw_t*    Straw_p, 
             const char* NextChunk_p)
{
  printf("[%08X] \tJumping to %s\n", Straw_p, NextChunk_p);
}


void
StrawLogYield(Straw_t*    Straw_p, 
              const char* NextChunk_p)
{
  printf("[%08X] \tYielding, then jumping to %s\n", Straw_p, NextChunk_p);
}


void
StrawLogTerminateStraw(Straw_t*    Straw_p, 
                       Straw_t*    TerminatedStraw_p, 
                       const char* NotifyChunk_p)
{
  printf("[%08X] \tTerminating straw %08X, then running %s\n", Straw_p, 
         TerminatedStraw_p, NotifyChunk_p);
}


void
StrawLogInit(Straw_t*    Straw_p,
             const char* ChunkName_p)
{
  printf("[%08X] (%s)\n", Straw_p, ChunkName_p);
}


void
StrawLogWaitSignal(Straw_t*                    Straw_p,
                   const StrawSignalMapping_t* SignalMap_p)
{
  if (SignalMap_p[1].Chunk == NULL) {
    printf("[%08X] \tWaiting for signal\n", Straw_p);
  }
  else {
    printf("[%08X] \tWaiting for signals\n", Straw_p);
  }
}


void
StrawLogTerminationAttempt(Straw_t* Straw_p)
{
  printf("[%08X] (Termination attempt)\n", Straw_p);
}

void
StrawLogTerminateChildStraw(Straw_t* Straw_p,
                            Straw_t* ChildStraw_p,
                            int      SubStraw)
{
  if (SubStraw) {
    printf("[%08X] \tTerminating sub-straw %08X first\n", Straw_p, ChildStraw_p);
  }
  else {
    printf("[%08X] \tTerminating child-straw %08X first\n", Straw_p, ChildStraw_p);
  }
}


void
StrawLogExit(Straw_t* Straw_p)
{
  printf("[%08X] \tStraw exited\n", Straw_p);
}


void
StrawLogDeleted(Straw_t* Straw_p)
{
  printf("[%08X] \tStraw deleted\n", Straw_p);
}


void
StrawLogRunCleanup(Straw_t*    Straw_p,
                   const char* CleanupChunk_p)
{
  printf("[%08X] \tRunning cleanup-chunk %s first\n", Straw_p, CleanupChunk_p);
}


void
StrawLogRunChild(Straw_t* Straw_p,
                 Straw_t* NewStraw_p)
{
  printf("[%08X] \tForking off child-straw %08X\n", Straw_p, NewStraw_p);
}


void
StrawLogRunSub(Straw_t*    Straw_p,
               Straw_t*    NewStraw_p,
               const char* ReturnChunk_p)
{
  printf("[%08X] \tRunning sub-straw %08X, returning to %s when done\n", Straw_p, NewStraw_p, 
         ReturnChunk_p);
}

#endif /* STRAW_CFG_DEBUG */
